/*
 * GnuAPDF Library
 *   Jon Sevy, jsevy@cs.drexel.edu
 * 
 * Based on GnuJPDF Library, authored by (among others)
 *   Peter T Mount, http://www.retep.org.uk/pdf/
 *   Eric Z. Beard, ericzbeard@hotmail.com
 *   Gilbert DeLeeuw, gil1@users.sourceforge.net
 *   Matthew Hreljac, mhreljac@hotmail.com  
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package gnu.apdf;


import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.Locale;
import java.util.Stack;

import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.DashPathEffect;
import android.graphics.DrawFilter;
import android.graphics.Matrix;
import android.graphics.Paint;
import android.graphics.Paint.Style;
import android.graphics.Path;
import android.graphics.PathEffect;
import android.graphics.PathMeasure;
import android.graphics.Picture;
import android.graphics.PointF;
import android.graphics.PorterDuff;
import android.graphics.PorterDuffXfermode;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.Region;
import android.graphics.Typeface;
import android.text.TextUtils;


/**
 * The PDFCanvas class holds the "draw" calls. To draw something, you use the standard Android Canvas drawing calls
 * (including drawing, transformations, clips, etc.) The PDFCanvas will encode these as PDF commands.
 * <p>
 * The PDFCanvas is associated with a PDFPage, which is part of a PDFDocument. The basic workflow is:
 * </p>
 * <code><pre>
 *      // Create a PDF document, a page in the document, and get that page's PDFCanvas
 *      PDFDocument pdfDoc = new PDFDocument();
 *      PDFPage page = new PDFPage(PageFormat.LETTER);
 *      pdfDoc.add(page);
 *      PDFCanvas pdfCanvas = page.getCanvas();
 *      
 *      // do drawing commands as on any other Canvas
 *      drawDesign(pdfCanvas);  
 *      
 *      // write the PDF content for the page to a file
 *      BufferedOutputStream fileStream = new BufferedOutputStream(new FileOutputStream(file));
 *      pdfDoc.write(outStream);
 *      fileStream.flush();
 *      fileStream.close();
 * </pre></code>
 * <p>
 * Note: Certain methods are unimplemented due to lack of support in PDF; these will throw an
 * UnsupportedOperationException if used. Some additional features currently have no or limited support, as indicated
 * below.
 * </p>
 * <p>
 * Supported operations:
    <ul>
      <li>Drawing operations, including </li>
      <ul>
        <li>drawLine</li>
        <li>drawArc</li>
        <li>drawCircle</li>
        <li>drawOval</li>
        <li>drawRect</li>
        <li>drawRoundRect</li>
        <li>drawPoints</li>
        <li>drawColor</li>
        <li>drawPaint</li>
      </ul>
      <li>Image drawing through drawBitmap</li>
      <li>Grid and 3D drawing through drawBitmapMesh and drawVertices</li>
      <li>Text operations drawText, drawPosText, drawTextOnPath</li>
      <li>Clipping through clipRect</li>
      <li>Matrix transformation operations including scale, skew, translate, rotate and concat</li>
      <li>State save and restore</li>
    </ul>
    </p>
    <p>
    Not supported:
    <ul>
      <li>Dotted lines</li>
      <li>Underlined and strikethrough text</li>
      <li>quickReject() methods (always return false)</li>
      <li>drawTextRun method (throws UnsupportedOperationException)</li>
      <li>clipPath method (throws UnsupportedOperationException)</li>
      <li>clipRect method with operation other than intersect (throws UnsupportedOperationException)</li>
      <li>clipRegion method (throws UnsupportedOperationException)</li>
      <li>drawPath method (throws UnsupportedOperationException)</li>
      <li>drawPicture method (throws UnsupportedOperationException)</li>
      <li>drawPatch method (throws UnsupportedOperationException)</li>
      <li>Paint xfermodes</li>
      <li>State-save flags and layer information are ignored; save(flags),
        saveLayer, and saveLayerAlpha all behave the same as plain
        save()</li>
    </ul>
 * </p>
 *
 * @author Jon Sevy, jsevy@cs.drexel.edu
 * 
 */
public class PDFCanvas extends Canvas
{
    
    /**
     * Constants for stroke cap operators
     */
    public static final int PDF_STROKE_CAP_BUTT = 0;
    public static final int PDF_STROKE_CAP_ROUND = 1;
    public static final int PDF_STROKE_CAP_PROJECTING_SQUARE = 2;
    
    /**
     * Constants for stroke join operators
     */
    public static final int PDF_STROKE_JOIN_MITER = 0;
    public static final int PDF_STROKE_JOIN_ROUND = 1;
    public static final int PDF_STROKE_JOIN_BEVEL = 2;
    
    /**
     * Page and page format for this graphics object
     */
    private PDFPage page;
    private PDFPageFormat pageFormat;
    
    /**
     * Current transform Matrix
     */
    private Matrix javaTransformMatrix = new Matrix();
    private Matrix javaToPDFGraphicsMatrix = new Matrix();
    private Matrix javaToPDFTextMatrix = new Matrix();
    private Matrix graphicsMatrix = new Matrix();
    private Matrix textMatrix = new Matrix();
    
    /**
     * Clip region - initially the largest rectangle
     */
    private Region clipRegion = new Region(Integer.MIN_VALUE, Integer.MIN_VALUE, Integer.MAX_VALUE, Integer.MAX_VALUE);
    
    /**
     * Class to use in graphics state save/restore stack; especially important because the PDF clip region can't be
     * expanded, only reduced...
     * 
     * @author jsevy
     * 
     */
    private class GraphicsState
    {
        Region clipRegion;
        Matrix javaTransformMatrix;
        
        
        
        public GraphicsState(Matrix matrix, Region region)
        {
            this.clipRegion = new Region(region);
            this.javaTransformMatrix = new Matrix(matrix);
        }
    }
    
    private Stack<GraphicsState> graphicsStateStack = new Stack<GraphicsState>();
    
    // make sure use the decimal format with "." as separator, in case that's not the case for the default locale
    private static final DecimalFormat df = new DecimalFormat("#.###", new DecimalFormatSymbols(Locale.US));
    

    
    /**
     * Current paint object - the last one used
     */
    private Paint currentPaint;
    
    /**
     * This is the PrintWriter used to write PDF drawing commands to the Stream
     */
    private PrintWriter pw;
    
    /**
     * @hide
     */
    protected int mDensity = Bitmap.DENSITY_NONE;
    
    /**
     * Flag for drawTextRun indicating left-to-right run direction.
     * 
     * @hide
     */
    public static final int DIRECTION_LTR = 0;
    
    /**
     * Flag for drawTextRun indicating right-to-left run direction.
     * 
     * @hide
     */
    public static final int DIRECTION_RTL = 1;
    
    // Maximum bitmap size
    private static final int MAXMIMUM_BITMAP_SIZE = 32766;
    
    
    
    /**
     * Construct a canvas for the specified page size
     * 
     * @param page
     */
    public PDFCanvas(PDFPage page)
    {
        // get our page info
        pageFormat = page.getPageFormat();
        
        this.page = page;
        
        // Now create a stream to store the graphics in
        PDFStream stream = new PDFStream();
        
        // To view detail in uncompressed format comment out the next line
        // stream.setDeflate(true);
        
        PDFDocument document = page.getPDFDocument();
        document.add(stream);
        page.add(stream);
        pw = stream.getWriter();
        
        // initially, we are limited to the page size
        clipRegion = new Region(page.getImageableArea());
        
        // user-defined Java transform - initialize to identity matrix
        javaTransformMatrix = new Matrix();
        
        // transform from Java Space to PDF Space
        // origin at bottom of page rather than at top, and y grows up rather
        // than down
        javaToPDFGraphicsMatrix = new Matrix();
        javaToPDFGraphicsMatrix.setScale(1f, -1f);
        javaToPDFGraphicsMatrix.postTranslate(pageFormat.getMargin(), pageFormat.getHeight() - pageFormat.getMargin());
        
        // same for text, but no vertical flip
        javaToPDFTextMatrix = new Matrix();
        javaToPDFTextMatrix.postTranslate(pageFormat.getMargin(), pageFormat.getHeight() - pageFormat.getMargin());
        
        // combine Java and Java2PDF matrices for overall transform
        setMatrix();
        
        // set colorspace to RGB
        // pw.println("/DeviceRGB CS");
        // pw.println("/DeviceRGB cs");
        
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException
     */
    public PDFCanvas(Bitmap bitmap)
    {
        throw new UnsupportedOperationException("Can't set a bitmap on a PDF canvas");
    }
    
    
    
    /**
     * Indicates whether this Canvas uses hardware acceleration - PDFCanvas does not.
     * 
     * @return Always false
     */
    public boolean isHardwareAccelerated()
    {
        return false;
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException
     */
    public void setBitmap(Bitmap bitmap)
    {
        throw new UnsupportedOperationException("Can't set a bitmap on a PDF canvas");
    }
    
    
    
    /**
     * Set the viewport dimensions if this canvas is GL based. If it is not, this method is ignored and no exception is
     * thrown. Since PDFCanvas is no GL-based, does nothing.
     * 
     * @param width
     *            The width of the viewport
     * @param height
     *            The height of the viewport
     * 
     * @hide
     */
    public void setViewport(int width, int height)
    {
    }
    
    
    
    /**
     * Return true if the device that the current layer draws into is opaque (i.e. does not support per-pixel alpha).
     * PDFCanvas only supports opaques layers, so always returns true.
     * 
     * @return Always returns true
     */
    public boolean isOpaque()
    {
        return true;
    }
    
    
    
    /**
     * Returns the imageable width of the current page
     * 
     * @return the imageable width of the current page
     */
    public int getWidth()
    {
        return pageFormat.getImageableWidth();
    }
    
    
    
    /**
     * Returns the imageable height of the current page
     * 
     * @return the imageable height of the current page
     */
    public int getHeight()
    {
        return pageFormat.getImageableHeight();
    }
    
    
    
    /**
     * <p>
     * Returns the target density of the canvas.
     * </p>
     * 
     * @return Returns the current target density of the canvas, which is used to determine the scaling factor when
     *         drawing a bitmap into it.
     * 
     * @see #setDensity(int)
     * @see Bitmap#getDensity()
     */
    public int getDensity()
    {
        return mDensity;
    }
    
    
    
    /**
     * <p>
     * Specifies the density for this Canvas' backing bitmap. This modifies the target density of the canvas itself, as
     * well as the density of its backing bitmap via {@link Bitmap#setDensity(int) Bitmap.setDensity(int)}.
     * 
     * @param density
     *            The new target density of the canvas, which is used to determine the scaling factor when drawing a
     *            bitmap into it. Use {@link Bitmap#DENSITY_NONE} to disable bitmap scaling.
     * 
     * @see #getDensity()
     * @see Bitmap#setDensity(int)
     */
    public void setDensity(int density)
    {
        mDensity = density;
    }
    
    
    
    /**
     * Not supported; ignored.
     * 
     * @param density
     */
    public void setScreenDensity(int density)
    {
        // does nothing
    }
    
    
    
    /**
     * Returns the maximum allowed width for bitmaps drawn with this canvas. Attempting to draw with a bitmap wider than
     * this value will result in an error.
     * 
     * @see #getMaximumBitmapHeight()
     */
    public int getMaximumBitmapWidth()
    {
        return MAXMIMUM_BITMAP_SIZE;
    }
    
    
    
    /**
     * Returns the maximum allowed height for bitmaps drawn with this canvas. Attempting to draw with a bitmap taller
     * than this value will result in an error.
     * 
     * @see #getMaximumBitmapWidth()
     */
    public int getMaximumBitmapHeight()
    {
        return MAXMIMUM_BITMAP_SIZE;
    }
    
    
    
    /**
     * Saves the current matrix and clip onto a private stack. Subsequent calls to translate,scale,rotate,skew,concat or
     * clipRect,clipPath will all operate as usual, but when the balancing call to restore() is made, those calls will
     * be forgotten, and the settings that existed before the save() will be reinstated.
     * 
     * @return The value to pass to restoreToCount() to balance this save()
     */
    public int save()
    {
        graphicsStateStack.push(new GraphicsState(javaTransformMatrix, clipRegion));
        pushState();
        return graphicsStateStack.size() - 1;
    }
    
    
    
    /*
     * Flag definitions for saving state // restore the current matrix when restore() is called public static final int
     * MATRIX_SAVE_FLAG = 0x01; // restore the current clip when restore() is called public static final int
     * CLIP_SAVE_FLAG = 0x02; // the layer needs to per-pixel alpha public static final int HAS_ALPHA_LAYER_SAVE_FLAG =
     * 0x04; // the layer needs to 8-bits per color component public static final int FULL_COLOR_LAYER_SAVE_FLAG = 0x08;
     * // clip against the layer's bounds public static final int CLIP_TO_LAYER_SAVE_FLAG = 0x10; // restore everything
     * when restore() is called public static final int ALL_SAVE_FLAG = 0x1F;
     */
    
    /**
     * Based on saveFlags, can save the current matrix and clip onto a private stack. Subsequent calls to
     * translate,scale,rotate,skew,concat or clipRect,clipPath will all operate as usual, but when the balancing call to
     * restore() is made, those calls will be forgotten, and the settings that existed before the save() will be
     * reinstated.
     * 
     * Currently behaves the same as save() - ignores saveFlags and simply saves the entire state.
     * 
     * @param saveFlags
     *            flag bits that specify which parts of the Canvas state to save/restore - currently ignored
     * @return The value to pass to restoreToCount() to balance this save()
     */
    public int save(int saveFlags)
    {
        return save();
    }
    
    
    
    /**
     * This behaves the same as save(), and all arguments are ignored.
     * 
     * @param bounds
     *            Ignored
     * @param paint
     *            Ignored
     * @param saveFlags
     *            Ignored
     * @return value to pass to restoreToCount() to balance this save()
     */
    public int saveLayer(RectF bounds, Paint paint, int saveFlags)
    {
        return save();
    }
    
    
    
    /**
     * This behaves the same as save(), and all arguments are ignored.
     */
    public int saveLayer(float left, float top, float right, float bottom, Paint paint, int saveFlags)
    {
        return save();
    }
    
    
    
    /**
     * This behaves the same as save(), and all arguments are ignored.
     * 
     * @param bounds
     *            Ignored
     * @param alpha
     *            Ignored
     * @param saveFlags
     *            Ignored
     * @return value to pass to restoreToCount() to balance this call
     */
    public int saveLayerAlpha(RectF bounds, int alpha, int saveFlags)
    {
        return save();
    }
    
    
    
    /**
     * This behaves the same as save(), and all arguments are ignored.
     */
    public int saveLayerAlpha(float left, float top, float right, float bottom, int alpha, int saveFlags)
    {
        return save();
    }
    
    
    
    /**
     * This call balances a previous call to save(), and is used to remove all modifications to the matrix/clip state
     * since the last save call. It is an error to call restore() more times than save() was called.
     */
    public void restore()
    {
        // will throw EmptyStackException if nothing to pop
        GraphicsState graphicsState = graphicsStateStack.pop();
        javaTransformMatrix = graphicsState.javaTransformMatrix;
        clipRegion = graphicsState.clipRegion;
        popState();
    }
    
    
    
    /**
     * Returns the number of matrix/clip states on the Canvas' private stack. This will equal # save() calls - #
     * restore() calls.
     */
    public int getSaveCount()
    {
        return graphicsStateStack.size();
    }
    
    
    
    /**
     * Efficient way to pop any calls to save() that happened after the save count reached saveCount. It is an error for
     * saveCount to be less than 1.
     * 
     * Example: int count = canvas.save(); ... // more calls potentially to save() ... canvas.restoreToCount(count); //
     * now the canvas is back in the same state it was before the initial call to save().
     * 
     * @param saveCount
     *            The save level to restore to.
     */
    public void restoreToCount(int saveCount)
    {
        // pop elements off stack until reach the specified saveCount
        // will throw EmptyStackException if not enough to pop
        GraphicsState graphicsState = graphicsStateStack.pop();
        popState();
        while (graphicsStateStack.size() > saveCount)
        {
            graphicsState = graphicsStateStack.pop();
            popState();
        }
        javaTransformMatrix = graphicsState.javaTransformMatrix;
        clipRegion = graphicsState.clipRegion;
    }
    
    
    
    @SuppressWarnings("unused")
    private void debugArray(String name, float[] array)
    {
        /* debugging stuff... */
        int length = array.length;
        pw.println("% " + name + ":");
        pw.print("% coeffs:     ");
        for (int i = 0; i < length; i++)
            pw.print(" " + df.format(array[i]));
        pw.println();
    }
    
    
    
    private void debugMatrix(String name, Matrix matrix)
    {
        /* debugging stuff... */
        float[] coef = new float[9];
        
        matrix.getValues(coef);
        pw.println("% " + name + ":");
        pw.println("% { " + "{ " + df.format(coef[Matrix.MSCALE_X]) + " " + df.format(coef[Matrix.MSKEW_X]) + " " + df.format(coef[Matrix.MTRANS_X]) + " }, " + "{ " + df.format(coef[Matrix.MSKEW_Y]) + " " + df.format(coef[Matrix.MSCALE_Y]) + " " + df.format(coef[Matrix.MTRANS_Y]) + " }, " + "{ 0 0 1 } }");
        
    }
    
    
    
    @SuppressWarnings("unused")
    private void debugMatrices()
    {
        /* debugging stuff... */
        debugMatrix("javaTransformMatrix", javaTransformMatrix);
        debugMatrix("graphicsMatrix", graphicsMatrix);
        debugMatrix("textMatrix", textMatrix);
    }
    
    
    
    /**
     * Create composite matrix from Java transform and Java-to-PDF transform
     */
    private void setMatrix()
    {
        // set both graphics and text matrices
        graphicsMatrix.setConcat(javaToPDFGraphicsMatrix, javaTransformMatrix);
        
        // text matrix is a bit odd... need to swap the x and y skews, negating
        // the new y skew and y scale...
        float[] coef = new float[9];
        graphicsMatrix.getValues(coef);
        float skewX = coef[Matrix.MSKEW_Y];
        float skewY = -coef[Matrix.MSKEW_X];
        float scaleY = -coef[Matrix.MSCALE_Y];
        coef[Matrix.MSKEW_X] = skewX;
        coef[Matrix.MSKEW_Y] = skewY;
        coef[Matrix.MSCALE_Y] = scaleY;
        textMatrix.setValues(coef);
        
    }
    
    
    
    /**
     * Set drawing characteristics as specified in new paint object where different from current, and store new Pint
     * object for use in stroking.
     * 
     * @param newPaint
     */
    private void setPaint(Paint newPaint)
    {
        
        // set drawing characteristics if changed
        
        // current drawing color
        if ((currentPaint == null) || (currentPaint.getColor() != newPaint.getColor()))
        {
            int color = newPaint.getColor();
            float redIntensity = ((float) Color.red(color)) / 255;
            float greenIntensity = ((float) Color.green(color)) / 255;
            float blueIntensity = ((float) Color.blue(color)) / 255;
            
            // set for both stroking and non-stroking operations
            pw.println(df.format(redIntensity) + " " + df.format(greenIntensity) + " " + df.format(blueIntensity) + " RG");
            pw.println(df.format(redIntensity) + " " + df.format(greenIntensity) + " " + df.format(blueIntensity) + " rg");
        }
        
        // line width
        if ((currentPaint == null) || (currentPaint.getStrokeWidth() != newPaint.getStrokeWidth()))
        {
            float strokeWidth = newPaint.getStrokeWidth();
            pw.println(df.format(strokeWidth) + " w");
        }
        
        // line end cap
        if ((currentPaint == null) || (currentPaint.getStrokeCap() != newPaint.getStrokeCap()))
        {
            Paint.Cap strokeCap = newPaint.getStrokeCap();
            int pdfStrokeCap;
            switch (strokeCap)
            {
                case ROUND:
                {
                    pdfStrokeCap = PDF_STROKE_CAP_ROUND;
                    break;
                }
                
                case SQUARE:
                {
                    pdfStrokeCap = PDF_STROKE_CAP_PROJECTING_SQUARE;
                    break;
                }
                
                case BUTT:
                default:
                {
                    pdfStrokeCap = PDF_STROKE_CAP_BUTT;
                    break;
                }
            }
            pw.println(pdfStrokeCap + " J");
        }
        
        // set stroke join style
        if ((currentPaint == null) || (currentPaint.getStrokeJoin() != newPaint.getStrokeJoin()))
        {
            Paint.Join strokeJoin = newPaint.getStrokeJoin();
            int pdfStrokeJoin;
            switch (strokeJoin)
            {
                case ROUND:
                {
                    pdfStrokeJoin = PDF_STROKE_JOIN_ROUND;
                    break;
                }
                
                case BEVEL:
                {
                    pdfStrokeJoin = PDF_STROKE_JOIN_BEVEL;
                    break;
                }
                
                case MITER:
                default:
                {
                    pdfStrokeJoin = PDF_STROKE_JOIN_MITER;
                    break;
                }
            }
            pw.println(pdfStrokeJoin + " j");
        }
        
        // set line join miter
        if ((currentPaint == null) || (currentPaint.getStrokeMiter() != newPaint.getStrokeMiter()))
        {
            float pdfMiterLimit = newPaint.getStrokeMiter();
            pw.println(df.format(pdfMiterLimit) + " M");
        }
        
        // To do - set line dash style...
        if ((currentPaint == null) || (currentPaint.getPathEffect() != newPaint.getPathEffect()))
        {
            PathEffect pathEffect = newPaint.getPathEffect();
            if (pathEffect == null)
            {
                // no path effect - make it a solid line
                pw.println("[] 0 d");
            }
            if (pathEffect instanceof DashPathEffect)
            {
                // set the dash pattern appropriately
                // DashPathEffect dashPathEffect = (DashPathEffect)(pathEffect);
                // EXCEPT no way to get hold of the intervals array or phase parameters... :-(
            }
        }
        
        // set font stuff - typeface and size
        if ((currentPaint == null) || ((currentPaint.getTypeface() != newPaint.getTypeface()) || (currentPaint.getTextSize() != newPaint.getTextSize())))
        {
            Typeface typeface = newPaint.getTypeface();
            if (typeface != null)
            {
                String fontName = this.page.getFont("/Type1", typeface).getName();
                int fontSize = (int) newPaint.getTextSize();
                pw.println(fontName + " " + fontSize + " Tf");
            }
        }
        
        // make a copy of the incoming paint object so changes there
        // don't automatically end up here...
        currentPaint = new Paint(newPaint);
        
    }
    
    
    
    /**
     * Append the correct operator to finish a sequence of path-painting operations
     */
    private void paintPath()
    {
        switch (currentPaint.getStyle())
        {
            case FILL:
            {
                pw.println("f");
            }
            
            case FILL_AND_STROKE:
            {
                pw.println("B");
            }
            
            case STROKE:
            default:
            {
                pw.println("S");
            }
        }
    }
    
    
    
    /**
     * Append the clipping operator after clip drawing operations
     */
    private void clipPath()
    {
        pw.println("W");
        pw.println("n");
    }
    
    
    
    /**
     * Used to eliminate small gap at end of e.g. circles
     */
    private void closePath()
    {
        pw.println("h");
    }
    
    
    
    /**
     * Start a text block, with text to be written at position (x, y)
     * 
     * @param x
     * @param y
     */
    private void startTextBlock(float x, float y)
    {
        // inherit the current text font and size from setting the paint object
        pw.println("BT");
        
        // apply current text transformation matrix, but with converted position
        // (x,y) as the translation
        float[] coef = new float[9];
        textMatrix.getValues(coef);
        pw.println(df.format(coef[Matrix.MSCALE_X]) + " " + df.format(coef[Matrix.MSKEW_X]) + " " + df.format(coef[Matrix.MSKEW_Y]) + " " + df.format(coef[Matrix.MSCALE_Y]) + " " + cxy(x, y) + "Tm");
        
    }
    
    
    
    /**
     * Append the end-text-block operator
     */
    private void endTextBlock()
    {
        pw.println("ET");
    }
    
    
    
    /**
     * Header for image data that includes appropriate transformation matrix, which includes existing Java
     * transformation matrix information but modifies scales to take into account the "unit square" interpretation of
     * PDF image space, and modifies the translations to account for the reversal of y axis.
     * 
     * @param x
     *            x coordinate of top left corner of image
     * @param y
     *            y coordinate of top left corner of image
     * @param width
     *            width of image
     * @param height
     *            heignt of image
     */
    private void startImage(float x, float y, float width, float height)
    {
        // save the transformation matrix as we're going to change it temporarily
        pw.println("q");
        
        // apply current text transformation matrix, but with converted position
        // (x,y) as the translation, and modified scales...
        Matrix imageMatrix = new Matrix(textMatrix);
        
        // need to scale the scales since the PDF image model assumes image fits
        // in a unit square; so multiply by the width and height
        imageMatrix.preScale(width, height);
        
        float[] coef = new float[9];
        imageMatrix.getValues(coef);
        
        // also need to shift down by the height since coords are for top left but
        // PDF uses bottom left
        y += height;
        
        pw.println(df.format(coef[Matrix.MSCALE_X]) + " " + df.format(coef[Matrix.MSKEW_X]) + " " + df.format(coef[Matrix.MSKEW_Y]) + " " + df.format(coef[Matrix.MSCALE_Y]) + " " + cxy(x, y) + "cm");
        
    }
    
    
    
    /**
     * Header for image data that includes appropriate transformation matrix, which includes existing Java
     * transformation matrix information but modifies scales to take into account the "unit square" interpretation of
     * PDF image space, and modifies the translations to account for the reversal of y axis.
     * 
     * * @param matrix Matrix to transform image location and appearance
     */
    private void startImage(Matrix matrix)
    {
        // save the transformation matrix as we're going to change it temporarily
        pw.println("q");
        
        float[] origCoef = new float[9];
        matrix.getValues(origCoef);
        
        // concat with text matrix to include scales, rotations and skews
        matrix.preConcat(textMatrix);
        // matrix.preConcat(graphicsMatrix);
        
        float[] concatCoef = new float[9];
        matrix.getValues(concatCoef);
        
        // we use the concatenated scales and skews, but the original translations mapped
        // by the graphics matrix...
        pw.println(df.format(concatCoef[Matrix.MSCALE_X]) + " " + df.format(concatCoef[Matrix.MSKEW_X]) + " " + df.format(concatCoef[Matrix.MSKEW_Y]) + " " + df.format(concatCoef[Matrix.MSCALE_Y]) + " " + cxy(origCoef[Matrix.MTRANS_X], origCoef[Matrix.MTRANS_Y]) + "cm");
        
    }
    
    
    
    /**
     * Restore transformation matrix saved when image block was started.
     */
    private void endImage()
    {
        // restore previous transformation matrix
        pw.println("Q");
    }
    
    
    
    /**
     * Push the current graphics state so can be restored with matching call to popState
     */
    private void pushState()
    {
        pw.println("q");
    }
    
    
    
    /**
     * Restore the graphics state
     */
    private void popState()
    {
        pw.println("Q");
        
        // PDF save/restore state is superset of Java save/restore; now need to restore
        // those graphics parameters that were changed by PDF restore but expected unchanged
        // by the Java restore; these include:
        
        // color, line width, line cap style, line join style, line join miter, line dash pattern,
        // font, font size
        // first set current Paint to null so will re-establish parameters in PDF output
        Paint paint = currentPaint;
        currentPaint = null;
        setPaint(paint);
        
    }
    
    
    
    /**
     * Preconcat the current matrix with the specified translation
     * 
     * @param dx
     *            The distance to translate in X
     * @param dy
     *            The distance to translate in Y
     */
    public void translate(float dx, float dy)
    {
        // pass to superclass so final method getMatrix returns correct result
        super.translate(dx, dy);
        javaTransformMatrix.preTranslate(dx, dy);
        setMatrix();
    }
    
    
    
    /**
     * Preconcat the current matrix with the specified scale.
     * 
     * @param sx
     *            The amount to scale in X
     * @param sy
     *            The amount to scale in Y
     */
    public void scale(float sx, float sy)
    {
        // pass to superclass so final method getMatrix returns correct result
        super.scale(sx, sy);
        javaTransformMatrix.preScale(sx, sy);
        setMatrix();
    }
    
    
    
    /*
     * Preconcat the current matrix with the specified scale.
     * 
     * @param sx The amount to scale in X
     * 
     * @param sy The amount to scale in Y
     * 
     * @param px The x-coord for the pivot point (unchanged by the scale)
     * 
     * @param py The y-coord for the pivot point (unchanged by the scale)
     */
    /*
     * public final void scale(float sx, float sy, float px, float py) { translate(px, py); scale(sx, sy);
     * translate(-px, -py); }
     */
    
    /**
     * Preconcat the current matrix with the specified rotation.
     * 
     * @param degrees
     *            The amount to rotate, in degrees
     */
    public void rotate(float degrees)
    {
        // pass to superclass so final method getMatrix returns correct result
        super.rotate(degrees);
        javaTransformMatrix.preRotate(degrees);
        setMatrix();
    }
    
    
    
    /*
     * Preconcat the current matrix with the specified rotation.
     * 
     * @param degrees The amount to rotate, in degrees
     * 
     * @param px The x-coord for the pivot point (unchanged by the rotation)
     * 
     * @param py The y-coord for the pivot point (unchanged by the rotation)
     */
    /*
     * public final void rotate(float degrees, float px, float py) { translate(px, py); rotate(degrees); translate(-px,
     * -py); }
     */
    
    /**
     * Preconcat the current matrix with the specified skew.
     * 
     * @param sx
     *            The amount to skew in X
     * @param sy
     *            The amount to skew in Y
     */
    public void skew(float sx, float sy)
    {
        // pass to superclass so final method getMatrix returns correct result
        super.skew(sx, sy);
        javaTransformMatrix.preSkew(sx, sy);
        setMatrix();
    }
    
    
    
    /**
     * Preconcat the current matrix with the specified matrix.
     * 
     * @param matrix
     *            The matrix to preconcatenate with the current matrix
     */
    public void concat(Matrix matrix)
    {
        // pass to superclass so final method getMatrix returns correct result
        super.concat(matrix);
        javaTransformMatrix.preConcat(matrix);
        setMatrix();
    }
    
    
    
    /**
     * Completely replace the current matrix with the specified matrix. If the matrix parameter is null, then the
     * current matrix is reset to identity.
     * 
     * @param matrix
     *            The matrix to replace the current matrix with. If it is null, set the current matrix to identity.
     */
    public void setMatrix(Matrix matrix)
    {
        // pass to superclass so final method getMatrix returns correct result
        super.setMatrix(matrix);
        javaTransformMatrix.set(matrix);
        setMatrix();
    }
    
    
    
    /**
     * Return, in ctm, the current transformation matrix. This does not alter the matrix in the canvas, but just returns
     * a copy of it.
     */
    public void getMatrix(Matrix ctm)
    {
        ctm.set(javaTransformMatrix);
    }
    
    
    
    /**
     * Return a new matrix with a copy of the canvas' current transformation matrix. Since this is declared final in
     * Canvas class, we can't override, but by using the base class setMatrix method our matrix will be returned by the
     * base class getMatrix method.
     */
    /*
     * public final Matrix getMatrix() { return new Matrix(currentMatrix); }
     */
    
    /**
     * Modify the current clip with the specified rectangle. Only supports INTERSECT as the operation; throws an
     * UnsupportedOperationException if another operation is supplied
     * 
     * @param rect
     *            The rect to intersect with the current clip
     * @param op
     *            How the clip is modified
     * @return true if the resulting clip is non-empty
     */
    public boolean clipRect(RectF rect, Region.Op op)
    {
        // if op isn't intersection, it's not supported - throw an exception
        if (op != Region.Op.INTERSECT)
            throw new UnsupportedOperationException("Operation " + op + " not supported in PDFCanvas clipRect function");
        
        // covert float rect to int rect
        return clipRect(new Rect((int) Math.round(rect.left), (int) Math.round(rect.top), (int) Math.round(rect.right), (int) Math.round(rect.bottom)), op);
    }
    
    
    
    /**
     * Modify the current clip with the specified rectangle. Only supports INTERSECT as the operation; throws an
     * UnsupportedOperationException if another operation is supplied
     * 
     * @param rect
     *            The rectangle to intersect with the current clip.
     * @param op
     *            How the clip is modified
     * @return true if the resulting clip is non-empty
     */
    public boolean clipRect(Rect rect, Region.Op op)
    {
        // if op isn't intersection, it's not supported - throw an exception
        if (op != Region.Op.INTERSECT)
            throw new UnsupportedOperationException("Operation " + op + " not supported in PDFCanvas clipRect function");
        
        // call superclass method to track clip region so final getClipBounds
        // method returns correct rect
        super.clipRect(rect, op);
        
        // update the clip region
        clipRegion.op(rect, op);
        
        // perform clip operations - clockwise traversal
        moveto(rect.left, rect.top);
        lineto(rect.right, rect.top);
        lineto(rect.right, rect.bottom);
        lineto(rect.left, rect.bottom);
        lineto(rect.left, rect.top);
        closePath();
        clipPath();
        
        // return whether the region is now empty
        return !clipRegion.isEmpty();
    }
    
    
    
    /**
     * Intersect the current clip with the specified rectangle, which is expressed in local coordinates.
     * 
     * @param rect
     *            The rectangle to intersect with the current clip.
     * @return true if the resulting clip is non-empty
     */
    public boolean clipRect(RectF rect)
    {
        return clipRect(rect, Region.Op.INTERSECT);
    }
    
    
    
    /**
     * Intersect the current clip with the specified rectangle, which is expressed in local coordinates.
     * 
     * @param rect
     *            The rectangle to intersect with the current clip.
     * @return true if the resulting clip is non-empty
     */
    public boolean clipRect(Rect rect)
    {
        return clipRect(rect, Region.Op.INTERSECT);
    }
    
    
    
    /**
     * Modify the current clip with the specified rectangle, which is expressed in local coordinates. Only supports
     * INTERSECT as the operation; throws an UnsupportedOperationException if another operation is supplied
     * 
     * @param left
     *            The left side of the rectangle to intersect with the current clip
     * @param top
     *            The top of the rectangle to intersect with the current clip
     * @param right
     *            The right side of the rectangle to intersect with the current clip
     * @param bottom
     *            The bottom of the rectangle to intersect with the current clip
     * @param op
     *            How the clip is modified
     * @return true if the resulting clip is non-empty
     */
    public boolean clipRect(float left, float top, float right, float bottom, Region.Op op)
    {
        return clipRect(new RectF(left, top, right, bottom), op);
    }
    
    
    
    /**
     * Intersect the current clip with the specified rectangle, which is expressed in local coordinates.
     * 
     * @param left
     *            The left side of the rectangle to intersect with the current clip
     * @param top
     *            The top of the rectangle to intersect with the current clip
     * @param right
     *            The right side of the rectangle to intersect with the current clip
     * @param bottom
     *            The bottom of the rectangle to intersect with the current clip
     * @return true if the resulting clip is non-empty
     */
    public boolean clipRect(float left, float top, float right, float bottom)
    {
        return clipRect(new RectF(left, top, right, bottom));
    }
    
    
    
    /**
     * Intersect the current clip with the specified rectangle, which is expressed in local coordinates.
     * 
     * @param left
     *            The left side of the rectangle to intersect with the current clip
     * @param top
     *            The top of the rectangle to intersect with the current clip
     * @param right
     *            The right side of the rectangle to intersect with the current clip
     * @param bottom
     *            The bottom of the rectangle to intersect with the current clip
     * @return true if the resulting clip is non-empty
     */
    public boolean clipRect(int left, int top, int right, int bottom)
    {
        return clipRect(new RectF(left, top, right, bottom));
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException if called.
     * 
     * @param path
     *            The path to operate on the current clip
     * @param op
     *            How the clip is modified
     * @return true if the resulting is non-empty
     */
    public boolean clipPath(Path path, Region.Op op)
    {
        throw new UnsupportedOperationException("Operation clipPath not currently supported in PDFCanvas");
        
        // call superclass method to track clip region so final getClipBounds
        // method returns correct rect
        // super.clipPath(path, op);
        // return true;
    }
    
    
    
    /**
     * Intersect the current clip with the specified path. Not supported; throws UnsupportedOperationException if
     * called.
     * 
     * @param path
     *            The path to intersect with the current clip
     * @return true if the resulting is non-empty
     */
    public boolean clipPath(Path path)
    {
        throw new UnsupportedOperationException("Operation clipPath not currently supported in PDFCanvas");
        
        // return clipPath(path, Region.Op.INTERSECT);
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException if called.
     * 
     * @param region
     *            The region to operate on the current clip, based on op
     * @param op
     *            How the clip is modified
     * @return true if the resulting is non-empty
     */
    public boolean clipRegion(Region region, Region.Op op)
    {
        throw new UnsupportedOperationException("clipRegion operation not supported in PDFCanvas");
        
        // if op isn't intersection, it's not supported - throw an exception
        // if (op != Region.Op.INTERSECT)
        // throw new UnsupportedOperationException("Operation " + op + " not supported in PDFCanvas clipRect function");
        
        // call superclass method to track clip region so final getClipBounds
        // method returns correct rect
        // super.clipPath(path, op);
        
        // update the clip region
        // clipRegion.op(region, op);
        
        // indicate if resulting region is empty
        // return !clipRegion.isEmpty();
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException if called.
     * 
     * @param region
     *            The region to operate on the current clip, based on op
     * @return true if the resulting is non-empty
     */
    public boolean clipRegion(Region region)
    {
        return clipRegion(region, Region.Op.INTERSECT);
    }
    
    
    
    /**
     * Set the clip to the specified polygonal region.
     * 
     * @param verts
     */
    private void clipPoly(float[] verts)
    {
        // only reasonable if have 3 or more points
        int numPoints = verts.length / 2;
        
        if (numPoints >= 3)
        {
            moveto(verts[0], verts[1]);
            
            for (int i = 1; i < numPoints; i++)
            {
                lineto(verts[2 * i], verts[2 * i + 1]);
            }
            
            // close it
            lineto(verts[0], verts[1]);
            
            clipPath();
        }
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException if called.
     */
    public DrawFilter getDrawFilter()
    {
        throw new UnsupportedOperationException("No DrawFilter associated with PDF canvas");
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException if called.
     */
    public void setDrawFilter(DrawFilter filter)
    {
        throw new UnsupportedOperationException("No DrawFilter associated with PDF canvas");
    }
    
    public enum EdgeType
    {
        BW(0), // !< treat edges by just rounding to nearest pixel boundary
        AA(1); // !< treat edges by rounding-out, since they may be antialiased
        
        EdgeType(int nativeInt)
        {
            this.nativeInt = nativeInt;
        }
        
        /**
         * @hide
         */
        public final int nativeInt;
    }
    
    
    
    /**
     * Not supported, so always returns false.
     * 
     * @param rect
     *            Ignored
     * @param type
     *            Ignored
     * @return Always returns false
     */
    public boolean quickReject(RectF rect, EdgeType type)
    {
        return false;
    }
    
    
    
    /**
     * Not supported, so always returns false.
     * 
     * @param path
     *            Ignored
     * @param type
     *            Ignored
     * @return Always returns false
     */
    public boolean quickReject(Path path, EdgeType type)
    {
        return false;
    }
    
    
    
    /**
     * Not supported, so always returns false.
     * 
     * @param left
     *            Ignored
     * @param top
     *            Ignored
     * @param right
     *            Ignored
     * @param bottom
     *            Ignored
     * @param type
     *            Ignored
     * @return Always returns false
     */
    public boolean quickReject(float left, float top, float right, float bottom, EdgeType type)
    {
        // To do
        return false;
    }
    
    
    
    /**
     * Retrieve the clip bounds, returning true if they are non-empty.
     * 
     * @param bounds
     *            Return the clip bounds here. If it is null, ignore it but still return true if the current clip is
     *            non-empty.
     * @return true if the current clip is non-empty.
     */
    public boolean getClipBounds(Rect bounds)
    {
        bounds = clipRegion.getBounds();
        return !clipRegion.isEmpty();
    }
    
    
    
    /**
     * Retrieve the clip bounds. Can't override the superclass method, so add superclass calls whenever managing the
     * clip.
     * 
     * @return the clip bounds, or [0, 0, 0, 0] if the clip is empty.
     */
    /*
     * public final Rect getClipBounds() { Rect r = new Rect(); getClipBounds(r); return r; }
     */
    
    /**
     * Fill the entire page (restricted to the current clip) with the specified RGB color.
     * 
     * @param r
     *            red component (0..255) of the color to draw onto the canvas
     * @param g
     *            green component (0..255) of the color to draw onto the canvas
     * @param b
     *            blue component (0..255) of the color to draw onto the canvas
     */
    public void drawRGB(int r, int g, int b)
    {
        int color = Color.rgb(r, g, b);
        Paint paint = new Paint();
        paint.setColor(color);
        paint.setStyle(Style.FILL);
        
        drawPaint(paint);
    }
    
    
    
    /**
     * Fill the entire page (restricted to the current clip) with the specified ARGB color.
     * 
     * @param a
     *            alpha component (0..255) of the color to draw onto the canvas
     * @param r
     *            red component (0..255) of the color to draw onto the canvas
     * @param g
     *            green component (0..255) of the color to draw onto the canvas
     * @param b
     *            blue component (0..255) of the color to draw onto the canvas
     */
    public void drawARGB(int a, int r, int g, int b)
    {
        int color = Color.argb(a, r, g, b);
        Paint paint = new Paint();
        paint.setColor(color);
        paint.setStyle(Style.FILL);
        
        drawPaint(paint);
    }
    
    
    
    /**
     * Fill the entire page (restricted to the current clip) with the specified color.
     * 
     * @param color
     *            the color to draw onto the canvas
     */
    public void drawColor(int color)
    {
        Paint paint = new Paint();
        paint.setColor(color);
        paint.setStyle(Style.FILL);
        
        drawPaint(paint);
    }
    
    
    
    /**
     * Fill the entire page (restricted to the current clip) with the specified color and porter-duff xfermode.
     * 
     * Xfermodes are not currently supported, so always opaque overwrite.
     * 
     * @param color
     *            the color to draw with
     * @param mode
     *            the porter-duff mode to apply to the color - currently ignored
     */
    public void drawColor(int color, PorterDuff.Mode mode)
    {
        Paint paint = new Paint();
        paint.setColor(color);
        paint.setXfermode(new PorterDuffXfermode(mode));
        drawPaint(paint);
    }
    
    
    
    /**
     * Fill the entire page (restricted to the current clip) with the specified paint.
     * 
     * @param paint
     *            The paint used to draw onto the canvas
     */
    public void drawPaint(Paint paint)
    {
        drawRect(0, 0, pageFormat.getImageableWidth(), pageFormat.getImageableHeight(), paint);
    }
    
    
    
    /**
     * Draw a series of points. Each point is centered at the coordinate specified by pts[], and its diameter is
     * specified by the paint's stroke width (as transformed by the canvas' CTM), with special treatment for a stroke
     * width of 0, which always draws exactly 1 pixel (or at most 4 if antialiasing is enabled). The shape of the point
     * is controlled by the paint's Cap type. The shape is a square, unless the cap type is Round, in which case the
     * shape is a circle.
     * 
     * @param pts
     *            Array of points to draw [x0 y0 x1 y1 x2 y2 ...]
     * @param offset
     *            Number of values to skip before starting to draw.
     * @param count
     *            The number of values to process, after skipping offset of them. Since one point uses two values, the
     *            number of "points" that are drawn is really (count >> 1).
     * @param paint
     *            The paint used to draw the points
     */
    public void drawPoints(float[] pts, int offset, int count, Paint paint)
    {
        // check to make sure we have enough elements in array; just punt if not
        if ((pts.length < offset + count) || (pts.length - offset < 2))
            return;
        
        int i = offset;
        while ((count - i) >= 2)
        {
            drawPoint(pts[i], pts[i + 1], paint);
            i += 2;
        }
    }
    
    
    
    /**
     * Helper for drawPoints() that assumes you want to draw the entire array
     */
    public void drawPoints(float[] pts, Paint paint)
    {
        drawPoints(pts, 0, pts.length, paint);
    }
    
    
    
    /**
     * Helper for drawPoints() for drawing a single point.
     */
    public void drawPoint(float x, float y, Paint paint)
    {
        setPaint(paint);
        
        // draw degenerate line; point won't show up unless cap is round (or square?)
        moveto(x, y);
        lineto(x, y);
        paintPath();
    }
    
    
    
    /**
     * Draw a line segment with the specified start and stop x,y coordinates, using the specified paint.
     * 
     * @param startX
     *            The x-coordinate of the start point of the line
     * @param startY
     *            The y-coordinate of the start point of the line
     * @param paint
     *            The paint used to draw the line
     */
    public void drawLine(float startX, float startY, float stopX, float stopY, Paint paint)
    {
        setPaint(paint);
        moveto(startX, startY);
        lineto(stopX, stopY);
        paintPath();
    }
    
    
    
    /**
     * Draw a series of lines. Each line is taken from 4 consecutive values in the pts array. Thus to draw 1 line, the
     * array must contain at least 4 values. This is logically the same as drawing the array as follows:
     * drawLine(pts[0], pts[1], pts[2], pts[3]) followed by drawLine(pts[4], pts[5], pts[6], pts[7]) and so on.
     * 
     * @param pts
     *            Array of points to draw [x0 y0 x1 y1 x2 y2 ...]
     * @param offset
     *            Number of values in the array to skip before drawing.
     * @param count
     *            The number of values in the array to process, after skipping "offset" of them. Since each line uses 4
     *            values, the number of "lines" that are drawn is really (count >> 2).
     * @param paint
     *            The paint used to draw the points
     */
    public void drawLines(float[] pts, int offset, int count, Paint paint)
    {
        // check to make sure we have enough elements in array; just punt if not
        if ((pts.length < offset + count) || (pts.length - offset < 4))
            return;
        
        setPaint(paint);
        
        int i = offset;
        while ((count - i) >= 4)
        {
            moveto(pts[i], pts[i + 1]);
            lineto(pts[i + 2], pts[i + 3]);
            i += 4;
        }
        
        paintPath();
    }
    
    
    
    /**
     * Helper method for drawLines for the common case where all points are to be used.
     */
    public void drawLines(float[] pts, Paint paint)
    {
        drawLines(pts, 0, pts.length, paint);
    }
    
    
    
    /**
     * Draw the specified Rect using the specified paint. The rectangle will be filled or framed based on the Style in
     * the paint.
     * 
     * @param rect
     *            The rect to be drawn
     * @param paint
     *            The paint used to draw the rect
     */
    public void drawRect(RectF r, Paint paint)
    {
        drawRect(r.left, r.top, r.right, r.bottom, paint);
    }
    
    
    
    /**
     * Draw the specified Rect using the specified Paint. The rectangle will be filled or framed based on the Style in
     * the paint.
     * 
     * @param r
     *            The rectangle to be drawn.
     * @param paint
     *            The paint used to draw the rectangle
     */
    public void drawRect(Rect r, Paint paint)
    {
        drawRect(r.left, r.top, r.right, r.bottom, paint);
    }
    
    
    
    /**
     * Draw the specified Rect using the specified paint. The rectangle will be filled or framed based on the Style in
     * the paint.
     * 
     * @param left
     *            The left side of the rectangle to be drawn
     * @param top
     *            The top side of the rectangle to be drawn
     * @param right
     *            The right side of the rectangle to be drawn
     * @param bottom
     *            The bottom side of the rectangle to be drawn
     * @param paint
     *            The paint used to draw the rect
     */
    public void drawRect(float left, float top, float right, float bottom, Paint paint)
    {
        setPaint(paint);
        
        // clockwise traversal
        moveto(left, top);
        lineto(right, top);
        lineto(right, bottom);
        lineto(left, bottom);
        lineto(left, top);
        closePath();
        paintPath();
    }
    
    
    
    /**
     * Draw the specified oval using the specified paint. The oval will be filled or framed based on the Style in the
     * paint.
     * 
     * @param oval
     *            The rectangle bounds of the oval to be drawn
     */
    public void drawOval(RectF oval, Paint paint)
    {
        if (oval == null)
        {
            throw new NullPointerException();
        }
        drawArc(oval, 0, 360, false, paint);
    }
    
    
    
    /**
     * Draw the specified circle using the specified paint. If radius is <= 0, then nothing will be drawn. The circle
     * will be filled or framed based on the Style in the paint.
     * 
     * @param cx
     *            The x-coordinate of the center of the cirle to be drawn
     * @param cy
     *            The y-coordinate of the center of the cirle to be drawn
     * @param radius
     *            The radius of the cirle to be drawn
     * @param paint
     *            The paint used to draw the circle
     */
    public void drawCircle(float cx, float cy, float radius, Paint paint)
    {
        RectF bounds = new RectF(cx - radius, cy - radius, cx + radius, cy + radius);
        drawOval(bounds, paint);
    }
    
    
    
    /**
     * <p>
     * Draw the specified arc segment, which will be scaled to fit inside the specified oval. Note that this does not
     * stroke the arc, so it can be used to define a path consisting of additional segments and curves before stroking.
     * It also expects as a precondition that the starting point has been set with moveto or the result of another
     * graphics operation - it does not set the corrdinates of the first point in the arc.
     * </p>
     * 
     * <p>
     * If the start angle is negative or >= 360, the start angle is treated as start angle modulo 360.
     * </p>
     * 
     * <p>
     * If the sweep angle is >= 360, then the oval is drawn completely. Note that this differs slightly from
     * SkPath::arcTo, which treats the sweep angle modulo 360. If the sweep angle is negative, the sweep angle is
     * treated as sweep angle modulo 360
     * </p>
     * 
     * <p>
     * The arc is drawn clockwise. An angle of 0 degrees correspond to the geometric angle of 0 degrees (3 o'clock on a
     * watch.)
     * </p>
     * 
     * @param oval
     *            The bounds of oval used to define the shape and size of the arc
     * @param startAngle
     *            Starting angle (in degrees) where the arc begins
     * @param sweepAngle
     *            Sweep angle (in degrees) measured clockwise
     * @param paint
     *            The paint used to draw the arc
     */
    private void addArcSegment(RectF oval, float startAngle, float sweepAngle, Paint paint)
    {
        if (oval == null)
        {
            throw new NullPointerException();
        }
        
        setPaint(paint);
        
        // break into multiple Bezier curves, using approximately 30 degree
        // segments
        float segmentAngle = 30; /* degrees */
        int numSegments = (int) Math.ceil(Math.abs(sweepAngle / segmentAngle));
        segmentAngle = (sweepAngle) / numSegments;
        
        // half-angle in radians; used to get fundamental bezier coords
        float theta = (float) ((segmentAngle / 2) * (Math.PI / 180));
        
        // get oval parameters
        float xCen = (oval.right + oval.left) / 2;
        float yCen = (oval.bottom + oval.top) / 2;
        float xRadius = (oval.right - oval.left) / 2;
        float yRadius = (oval.bottom - oval.top) / 2;
        
        // get fundamental coords for control points P1, P2 for unit circle arc
        // from -theta degrees to theta degrees
        // these will be rotated and scaled and translated for each arc segment
        float p3x = (float) Math.cos(theta);
        float p3y = (float) Math.sin(theta);
        float p2x = (float) (4 - Math.cos(theta)) / 3;
        float p2y = (float) (((1 - Math.cos(theta)) * (3 - Math.cos(theta))) / (3 * Math.sin(theta)));
        float p1x = p2x;
        float p1y = -p2y;
        float p0x = p3x;
        float p0y = -p3y;
        float[] baseControlPoints = { p0x, p0y, p1x, p1y, p2x, p2y, p3x, p3y };
        float[] mappedControlPoints = { p0x, p0y, p1x, p1y, p2x, p2y, p3x, p3y };
        
        for (int i = 0; i < numSegments; i++)
        {
            float segmentStartAngle = startAngle + i * segmentAngle;
            
            // rotate, scale, and translate the base control points into the
            // mapped ones
            Matrix mat = new Matrix();
            mat.setRotate(segmentStartAngle + segmentAngle / 2);
            mat.postScale(xRadius, yRadius);
            mat.postTranslate(xCen, yCen);
            mat.mapPoints(mappedControlPoints, baseControlPoints);
            
            // define mapped points for pure lexical convenience; have to assume
            // the compiler is smart enough to ignore this
            // float q0x = mappedControlPoints[0];
            // float q0y = mappedControlPoints[1];
            float q1x = mappedControlPoints[2];
            float q1y = mappedControlPoints[3];
            float q2x = mappedControlPoints[4];
            float q2y = mappedControlPoints[5];
            float q3x = mappedControlPoints[6];
            float q3y = mappedControlPoints[7];
            
            // draw the Bezier segment for these control points
            // moveto(q0x, q0y);
            curveto(q1x, q1y, q2x, q2y, q3x, q3y);
        }
        
    }
    
    
    
    /**
     * For use in conjunction with addArcSegment when it is needed standalone rather than as an extension of a path in
     * progress; moves to the start of the arc segment
     * 
     * @param oval
     *            The bounds of oval used to define the shape and size of the arc
     * @param startAngle
     *            Starting angle (in degrees) where the arc begins
     */
    private void moveToArcStart(RectF oval, float startAngle)
    {
        // get oval parameters
        float xCen = (oval.right + oval.left) / 2;
        float yCen = (oval.bottom + oval.top) / 2;
        float xRadius = (oval.right - oval.left) / 2;
        float yRadius = (oval.bottom - oval.top) / 2;
        float startAngleRadians = startAngle * (float) (Math.PI / 180);
        
        float xStart = xCen + xRadius * (float) Math.cos(startAngleRadians);
        float yStart = yCen + yRadius * (float) Math.sin(startAngleRadians);
        
        moveto(xStart, yStart);
    }
    
    
    
    /**
     * <p>
     * Draw the specified arc, which will be scaled to fit inside the specified oval.
     * </p>
     * 
     * <p>
     * If the start angle is negative or >= 360, the start angle is treated as start angle modulo 360.
     * </p>
     * 
     * <p>
     * If the sweep angle is >= 360, then the oval is drawn completely. Note that this differs slightly from
     * SkPath::arcTo, which treats the sweep angle modulo 360. If the sweep angle is negative, the sweep angle is
     * treated as sweep angle modulo 360
     * </p>
     * 
     * <p>
     * The arc is drawn clockwise. An angle of 0 degrees correspond to the geometric angle of 0 degrees (3 o'clock on a
     * watch.)
     * </p>
     * 
     * @param oval
     *            The bounds of oval used to define the shape and size of the arc
     * @param startAngle
     *            Starting angle (in degrees) where the arc begins
     * @param sweepAngle
     *            Sweep angle (in degrees) measured clockwise
     * @param useCenter
     *            If true, include the center of the oval in the arc, and close it if it is being stroked. This will
     *            draw a wedge
     * @param paint
     *            The paint used to draw the arc
     */
    public void drawArc(RectF oval, float startAngle, float sweepAngle, boolean useCenter, Paint paint)
    {
        if (sweepAngle > 360)
            sweepAngle = 360;
        
        // if (sweepAngle < 0)
        // sweepAngle = (sweepAngle % 360) + 360; // mods of negatives give
        // negatives...
        
        // need to move to start point as addArcSegment expects this as
        // precondition
        moveToArcStart(oval, startAngle);
        addArcSegment(oval, startAngle, sweepAngle, paint);
        
        // add this to eliminate gap at end of full circle
        if (sweepAngle == 360)
            closePath();
        
        paintPath();
    }
    
    
    
    /**
     * Draw the specified round-rect using the specified paint. The roundrect will be filled or framed based on the
     * Style in the paint.
     * 
     * @param rect
     *            The rectangular bounds of the roundRect to be drawn
     * @param rx
     *            The x-radius of the oval used to round the corners
     * @param ry
     *            The y-radius of the oval used to round the corners
     * @param paint
     *            The paint used to draw the roundRect
     */
    public void drawRoundRect(RectF rect, float rx, float ry, Paint paint)
    {
        if (rect == null)
        {
            throw new NullPointerException();
        }
        
        setPaint(paint);
        
        // ick: draw ovals in corners connected by lines, clockwise from top
        // left
        
        // top line and top-right oval segment
        moveto(rect.left + rx, rect.top);
        lineto(rect.right - rx, rect.top);
        addArcSegment(new RectF(rect.right - 2 * rx, rect.top, rect.right, rect.top + 2 * ry), 270, 90, paint);
        
        // right line and bottom-right oval segment
        // moveto(rect.right, rect.top+ry);
        lineto(rect.right, rect.bottom - ry);
        addArcSegment(new RectF(rect.right - 2 * rx, rect.bottom - 2 * ry, rect.right, rect.bottom), 0, 90, paint);
        
        // bottom line and bottom-left oval segment
        // moveto(rect.right-rx, rect.bottom);
        lineto(rect.left + rx, rect.bottom);
        addArcSegment(new RectF(rect.left, rect.bottom - 2 * ry, rect.left + 2 * rx, rect.bottom), 90, 90, paint);
        
        // left line and top-left arc
        // moveto(rect.left, rect.bottom-ry);
        lineto(rect.left, rect.top + ry);
        addArcSegment(new RectF(rect.left, rect.top, rect.left + 2 * rx, rect.top + 2 * ry), 180, 90, paint);
        
        closePath();
        paintPath();
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException if called.
     * 
     * @param path
     *            The path to be drawn
     * @param paint
     *            The paint used to draw the path
     */
    public void drawPath(Path path, Paint paint)
    {
        // paths not supported at this point...
        throw new UnsupportedOperationException("Operation drawPath not currently supported in PDFCanvas");
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException if called.
     * 
     * @param bitmap
     *            The bitmap to draw as an N-patch
     * @param chunks
     *            The patches information (matches the native struct Res_png_9patch)
     * @param dst
     *            The destination rectangle.
     * @param paint
     *            The paint to draw the bitmap with. may be null
     * 
     * @hide
     */
    public void drawPatch(Bitmap bitmap, byte[] chunks, RectF dst, Paint paint)
    {
        throw new UnsupportedOperationException("Operation drawPatch not currently supported in PDFCanvas");
    }
    
    
    
    /**
     * Draw the specified bitmap, with its top/left corner at (x,y), using the specified paint, transformed by the
     * current matrix.
     * 
     * <p>
     * Note: if the paint contains a maskfilter that generates a mask which extends beyond the bitmap's original
     * width/height (e.g. BlurMaskFilter), then the bitmap will be drawn as if it were in a Shader with CLAMP mode. Thus
     * the color outside of the original width/height will be the edge color replicated.
     * 
     * <p>
     * If the bitmap and canvas have different densities, this function will take care of automatically scaling the
     * bitmap to draw at the same density as the canvas.
     * 
     * @param bitmap
     *            The bitmap to be drawn
     * @param left
     *            The position of the left side of the bitmap being drawn
     * @param top
     *            The position of the top side of the bitmap being drawn
     * @param paint
     *            The paint used to draw the bitmap (may be null)
     */
    public void drawBitmap(Bitmap bitmap, float left, float top, Paint paint)
    {
        PDFImage image = new PDFImage(bitmap);
        
        // Add the image to the document and the page resource-reference table
        page.getPDFDocument().add(image);
        page.addImageResource(image);
        page.addToProcset("/ImageC");
        
        startImage(left, top, bitmap.getWidth(), bitmap.getHeight());
        pw.println(image.getName() + " Do");
        endImage();
    }
    
    
    
    /**
     * Draw the specified bitmap, scaling/translating automatically to fill the destination rectangle. If the source
     * rectangle is not null, it specifies the subset of the bitmap to draw.
     * 
     * <p>
     * Note: if the paint contains a maskfilter that generates a mask which extends beyond the bitmap's original
     * width/height (e.g. BlurMaskFilter), then the bitmap will be drawn as if it were in a Shader with CLAMP mode. Thus
     * the color outside of the original width/height will be the edge color replicated.
     * 
     * <p>
     * This function <em>ignores the density associated with the bitmap</em>. This is because the source and destination
     * rectangle coordinate spaces are in their respective densities, so must already have the appropriate scaling
     * factor applied.
     * 
     * @param bitmap
     *            The bitmap to be drawn
     * @param src
     *            May be null. The subset of the bitmap to be drawn
     * @param dst
     *            The rectangle that the bitmap will be scaled/translated to fit into
     * @param paint
     *            May be null. The paint used to draw the bitmap
     */
    public void drawBitmap(Bitmap bitmap, Rect src, RectF dst, Paint paint)
    {
        
        PDFImage image;
        
        if (dst == null)
        {
            throw new NullPointerException();
        }
        
        if (src != null)
        {
            image = new PDFImage(bitmap, src.left, src.top, src.width(), src.height());
        }
        else
        {
            image = new PDFImage(bitmap);
        }
        
        // Add the image to the document and the page resource-reference table
        page.getPDFDocument().add(image);
        page.addImageResource(image);
        page.addToProcset("/ImageC");
        
        startImage(dst.left, dst.top, dst.width(), dst.height());
        pw.println(image.getName() + " Do");
        endImage();
    }
    
    
    
    /**
     * Draw the specified bitmap, scaling/translating automatically to fill the destination rectangle. If the source
     * rectangle is not null, it specifies the subset of the bitmap to draw.
     * 
     * <p>
     * Note: if the paint contains a maskfilter that generates a mask which extends beyond the bitmap's original
     * width/height (e.g. BlurMaskFilter), then the bitmap will be drawn as if it were in a Shader with CLAMP mode. Thus
     * the color outside of the original width/height will be the edge color replicated.
     * 
     * <p>
     * This function <em>ignores the density associated with the bitmap</em>. This is because the source and destination
     * rectangle coordinate spaces are in their respective densities, so must already have the appropriate scaling
     * factor applied.
     * 
     * @param bitmap
     *            The bitmap to be drawn
     * @param src
     *            May be null. The subset of the bitmap to be drawn
     * @param dst
     *            The rectangle that the bitmap will be scaled/translated to fit into
     * @param paint
     *            May be null. The paint used to draw the bitmap
     */
    public void drawBitmap(Bitmap bitmap, Rect src, Rect dst, Paint paint)
    {
        if (dst == null)
        {
            throw new NullPointerException();
        }
        
        drawBitmap(bitmap, src, new RectF(dst), paint);
    }
    
    
    
    /**
     * Treat the specified array of colors as a bitmap, and draw it. This gives the same result as first creating a
     * bitmap from the array, and then drawing it, but this method avoids explicitly creating a bitmap object which can
     * be more efficient if the colors are changing often.
     * 
     * @param colors
     *            Array of colors representing the pixels of the bitmap
     * @param offset
     *            Offset into the array of colors for the first pixel
     * @param stride
     *            The number of colors in the array between rows (must be >= width or <= -width).
     * @param x
     *            The X coordinate for where to draw the bitmap
     * @param y
     *            The Y coordinate for where to draw the bitmap
     * @param width
     *            The width of the bitmap
     * @param height
     *            The height of the bitmap
     * @param hasAlpha
     *            True if the alpha channel of the colors contains valid values. If false, the alpha byte is ignored
     *            (assumed to be 0xFF for every pixel).
     * @param paint
     *            May be null. The paint used to draw the bitmap
     */
    public void drawBitmap(int[] colors, int offset, int stride, float x, float y, int width, int height, boolean hasAlpha, Paint paint)
    {
        // check for valid input
        if (width < 0)
        {
            throw new IllegalArgumentException("width must be >= 0");
        }
        if (height < 0)
        {
            throw new IllegalArgumentException("height must be >= 0");
        }
        if (Math.abs(stride) < width)
        {
            throw new IllegalArgumentException("abs(stride) must be >= width");
        }
        int lastScanline = offset + (height - 1) * stride;
        int length = colors.length;
        if (offset < 0 || (offset + width > length) || lastScanline < 0 || (lastScanline + width > length))
        {
            throw new ArrayIndexOutOfBoundsException();
        }
        // quick escape if there's nothing to draw
        if (width == 0 || height == 0)
        {
            return;
        }
        
        // Be lazy - just create the damn Bitmap...
        Bitmap bitmap = Bitmap.createBitmap(colors, offset, stride, width, height, Bitmap.Config.ARGB_8888);
        drawBitmap(bitmap, x, y, paint);
    }
    
    
    
    /**
     * Legacy version of drawBitmap(int[] colors, ...) that took ints for x,y
     */
    public void drawBitmap(int[] colors, int offset, int stride, int x, int y, int width, int height, boolean hasAlpha, Paint paint)
    {
        // call through to the common float version
        drawBitmap(colors, offset, stride, (float) x, (float) y, width, height, hasAlpha, paint);
    }
    
    
    
    /**
     * Draw the bitmap using the specified matrix.
     * 
     * @param bitmap
     *            The bitmap to draw
     * @param matrix
     *            The matrix used to transform the bitmap when it is drawn
     * @param paint
     *            May be null. The paint used to draw the bitmap
     */
    public void drawBitmap(Bitmap bitmap, Matrix transformMatrix, Paint paint)
    {
        PDFImage image = new PDFImage(bitmap);
        
        // Add the image to the document and the page resource-reference table
        page.getPDFDocument().add(image);
        page.addImageResource(image);
        page.addToProcset("/ImageC");
        
        // create matrix copy since we're going to change it
        Matrix matrix = new Matrix(transformMatrix);
        
        // need to shift down by the bitmap height since Java coords are for top left but
        // PDF uses bottom left
        matrix.preTranslate(0, bitmap.getHeight());
        
        // need to scale since the PDF image model assumes image fits
        // in a unit square; so multiply by the width and height
        matrix.preScale(bitmap.getWidth(), bitmap.getHeight());
        
        // startImage(matrix, bitmap.getWidth(), bitmap.getHeight());
        startImage(matrix);
        pw.println(image.getName() + " Do");
        endImage();
    }
    
    
    
    protected static void checkRange(int length, int offset, int count)
    {
        if ((offset | count) < 0 || offset + count > length)
        {
            throw new ArrayIndexOutOfBoundsException();
        }
    }
    
    
    
    /**
     * Draw the bitmap through the mesh, where mesh vertices are evenly distributed across the bitmap. There are
     * meshWidth+1 vertices across, and meshHeight+1 vertices down. The verts array is accessed in row-major order, so
     * that the first meshWidth+1 vertices are distributed across the top of the bitmap from left to right.
     * 
     * This method stretches the supplied bitmap across the specified mesh, which can be used to create 3D effects in
     * which the bitmap appears to be applied to the surface for which the mesh is the projection on the plane of the
     * page.
     * 
     * @param bitmap
     *            The bitmap to draw using the mesh
     * @param meshWidth
     *            The number of columns in the mesh. Nothing is drawn if this is 0
     * @param meshHeight
     *            The number of rows in the mesh. Nothing is drawn if this is 0
     * @param verts
     *            Array of x,y pairs, specifying where the mesh should be drawn. There must be at least (meshWidth+1) *
     *            (meshHeight+1) * 2 + meshOffset values in the array
     * @param vertOffset
     *            Number of verts elements to skip before drawing
     * @param colors
     *            May be null. Specifies a color at each vertex, which is interpolated across the cell, and whose values
     *            are multiplied by the corresponding bitmap colors. If not null, there must be at least (meshWidth+1) *
     *            (meshHeight+1) + colorOffset values in the array.
     * @param colorOffset
     *            Number of color elements to skip before drawing
     * @param paint
     *            May be null. The paint used to draw the bitmap
     */
    public void drawBitmapMesh(Bitmap bitmap, int meshWidth, int meshHeight, float[] verts, int vertOffset, int[] colors, int colorOffset, Paint paint)
    {
        if ((meshWidth | meshHeight | vertOffset | colorOffset) < 0)
        {
            throw new ArrayIndexOutOfBoundsException();
        }
        if (meshWidth == 0 || meshHeight == 0)
        {
            return;
        }
        int count = (meshWidth + 1) * (meshHeight + 1);
        // we mul by 2 since we need two floats per vertex
        checkRange(verts.length, vertOffset, count * 2);
        if (colors != null)
        {
            // no mul by 2, since we need only 1 color per vertex
            checkRange(colors.length, colorOffset, count);
        }
        
        // OK get clever - draw a series of bitmaps for the triangles of the mesh,
        // by drawing them as clipped paralellograms
        for (int i = 0; i < meshWidth; i++)
        {
            for (int j = 0; j < meshHeight; j++)
            {
                int topLeftIndex = vertOffset + 2 * (i + j * (meshWidth + 1));
                int topRightIndex = vertOffset + 2 * ((i + 1) + j * (meshWidth + 1));
                int bottomLeftIndex = vertOffset + 2 * (i + (j + 1) * (meshWidth + 1));
                int bottomRightIndex = vertOffset + 2 * ((i + 1) + (j + 1) * (meshWidth + 1));
                
                // the points in our quadrilateral
                PointF topLeft = new PointF(verts[topLeftIndex], verts[topLeftIndex + 1]);
                PointF topRight = new PointF(verts[topRightIndex], verts[topRightIndex + 1]);
                PointF bottomLeft = new PointF(verts[bottomLeftIndex], verts[bottomLeftIndex + 1]);
                PointF bottomRight = new PointF(verts[bottomRightIndex], verts[bottomRightIndex + 1]);
                
                // subbitmap image
                int subbitmapWidth = bitmap.getWidth() / meshWidth;
                int subbitmapHeight = bitmap.getHeight() / meshHeight;
                PDFImage image = new PDFImage(bitmap, i * subbitmapWidth, j * subbitmapHeight, subbitmapWidth, subbitmapHeight);
                
                // Add the image to the document and the page resource-reference table
                page.getPDFDocument().add(image);
                page.addImageResource(image);
                page.addToProcset("/ImageC");
                
                // now create matrix to draw this bitmap into parallelogram containing top triangle
                Matrix matrix = new Matrix();
                
                // float[] topSrcBasis = {1, 0, 0, 1};
                // float[] topDstBasis = {bottomRight.x - bottomLeft.x, bottomRight.y - bottomLeft.y, topLeft.x -
                // bottomLeft.x, topLeft.y - bottomLeft.y};
                // debugArray("topSrcBasis", topSrcBasis);
                // debugArray("topDstBasis", topDstBasis);
                float[] topDstPoints = { topLeft.x, topLeft.y, topRight.x, topRight.y, bottomRight.x, bottomRight.y };
                
                // set matrix coefficients based on basis vectors and base point for translation; note that base point
                // is the completion of the top-right parallelogram - it's not one of the 4 points
                float[] matrixCoefs = new float[9];
                matrixCoefs[Matrix.MSCALE_X] = (topRight.x - topLeft.x);
                matrixCoefs[Matrix.MSKEW_X] = -(topRight.y - topLeft.y);
                matrixCoefs[Matrix.MSCALE_Y] = -(topRight.y - bottomRight.y);
                matrixCoefs[Matrix.MSKEW_Y] = (topRight.x - bottomRight.x);
                matrixCoefs[Matrix.MTRANS_X] = topLeft.x + (bottomRight.x - topRight.x);
                matrixCoefs[Matrix.MTRANS_Y] = topLeft.y + (bottomRight.y - topRight.y);
                matrixCoefs[Matrix.MPERSP_0] = 0;
                matrixCoefs[Matrix.MPERSP_1] = 0;
                matrixCoefs[Matrix.MPERSP_2] = 1;
                
                matrix.setValues(matrixCoefs);
                
                // draw the image, but clipping before to get just top right triangle;
                // save the state so we can undo the clipping
                save();
                clipPoly(topDstPoints);
                startImage(matrix);
                pw.println(image.getName() + " Do");
                endImage();
                restore();
                
                // now set matrix to draw this bitmap into parallelogram containing bottom-right triangle
                
                // we use the mapping of the fundamental unit basis to the sides of the triangle to create the matrix
                // scaling/skewing coefficients, and the location of the bottom-left point for the translation
                // float[] bottomSrcBasis = {1, 0, 0, 1};
                // float[] bottomDstBasis = {bottomRight.x - bottomLeft.x, bottomRight.y - bottomLeft.y, topLeft.x -
                // bottomLeft.x, topLeft.y - bottomLeft.y};
                // debugArray("bottomSrcBasis", bottomSrcBasis);
                // debugArray("bottomDstBasis", bottomDstBasis);
                float[] bottomDstPoints = { bottomLeft.x, bottomLeft.y, topLeft.x, topLeft.y, bottomRight.x, bottomRight.y };
                
                matrixCoefs = new float[9];
                matrixCoefs[Matrix.MSCALE_X] = (bottomRight.x - bottomLeft.x);
                matrixCoefs[Matrix.MSKEW_X] = -(bottomRight.y - bottomLeft.y);
                matrixCoefs[Matrix.MSCALE_Y] = -(topLeft.y - bottomLeft.y);
                matrixCoefs[Matrix.MSKEW_Y] = (topLeft.x - bottomLeft.x);
                matrixCoefs[Matrix.MTRANS_X] = bottomLeft.x;
                matrixCoefs[Matrix.MTRANS_Y] = bottomLeft.y;
                matrixCoefs[Matrix.MPERSP_0] = 0;
                matrixCoefs[Matrix.MPERSP_1] = 0;
                matrixCoefs[Matrix.MPERSP_2] = 1;
                
                matrix.setValues(matrixCoefs);
                
                // draw the image, but clipping before to get just bottom left triangle;
                // save the state so we can restore the clip
                save();
                clipPoly(bottomDstPoints);
                startImage(matrix);
                pw.println(image.getName() + " Do");
                endImage();
                restore();
                
            }
        }
        
    }
    
    
    
    /**
     * The VertexMode determines how the subsequent points in the sequence of vertices are interpreted.
     * 
     * - TRIANGLES In this mode each triple of points (6 successive coordinates) is treated as a separate triangle. This
     * is the most general mode; no assumption is made about triangle adjacency.
     * 
     * - TRIANGLE_STRIP In this mode each successive vertex is combined with the two before it to form a new triangle,
     * giving a "strip" of triangles where each shares an edge with the triangle before it
     * 
     * - TRIANGLE_FAN In this mode the first point is assumed to be the center vertex for all of the triangles, with
     * subsequent points defining the "outer" vertices of the fan arranged radially around the center. The triangles are
     * thus "pie wedges" with the first point at the center of the pie, and each "wedge" sharing an edge and vertex.
     * Note that the first and last outer vertices are not automatically joined.
     * 
     * @author jsevy
     *
     */
    /*
     * public enum VertexMode { TRIANGLES(0), TRIANGLE_STRIP(1), TRIANGLE_FAN(2);
     * 
     * VertexMode(int nativeInt) { this.nativeInt = nativeInt; }
     * 
     * public final int nativeInt; }
     */
    
    /**
     * Draw the array of vertices, interpreted as triangles (based on mode). The verts array is required, and specifies
     * the x,y pairs for each vertex. The color array specifies the color at each vertex, and the colors at each vertex
     * are interpolated across the corresponding triangle in a gradient. The indices and texs arrays are currently
     * ignored.
     * <p>
     * The VertexMode determines how the subsequent points in the sequence of vertices are interpreted.
     * <ul>
     * <li>
     * TRIANGLES In this mode each triple of points (6 successive coordinates) is treated as a separate triangle. This
     * is the most general mode; no assumption is made about triangle adjacency.</li>
     * <li>
     * TRIANGLE_STRIP In this mode each successive vertex is combined with the two before it to form a new triangle,
     * giving a "strip" of triangles where each shares an edge with the triangle before it</li>
     * <li>
     * TRIANGLE_FAN In this mode the first point is assumed to be the center vertex for all of the triangles, with
     * subsequent points defining the "outer" vertices of the fan arranged radially around the center. The triangles are
     * thus "pie wedges" with the first point at the center of the pie, and each "wedge" sharing an edge and vertex.
     * Note that the first and last outer vertices are not automatically joined.</li>
     * </ul>
     * </p>
     * 
     * @param mode
     *            How to interpret the array of vertices
     * @param vertexCount
     *            The number of values in the vertices array (and corresponding texs and colors arrays if non-null).
     *            Each logical vertex is two values (x, y), vertexCount must be a multiple of 2.
     * @param verts
     *            Array of vertices for the mesh
     * @param vertOffset
     *            Number of values in the verts to skip before drawing.
     * @param texs
     *            Currently ignored (may be null)
     * @param texOffset
     *            Currently ignored
     * @param colors
     *            Specifies a color for each vertex, to be interpolated across the triangle.
     * @param colorOffset
     *            Number of values in colors to skip before drawing.
     * @param indices
     *            Currently ignored (may be null)
     * @param indexCount
     *            Currently ignored
     * @param paint
     *            Currently ignored (may be null)
     */
    public void drawVertices(Canvas.VertexMode mode, int vertexCount, float[] verts, int vertOffset, float[] texs, int texOffset, int[] colors, int colorOffset, short[] indices, int indexOffset, int indexCount, Paint paint)
    {
        checkRange(verts.length, vertOffset, vertexCount);
        if (texs != null)
        {
            checkRange(texs.length, texOffset, vertexCount);
        }
        if (colors != null)
        {
            checkRange(colors.length, colorOffset, vertexCount / 2);
        }
        if (indices != null)
        {
            checkRange(indices.length, indexOffset, indexCount);
        }
        
        // convert vertices to pdf coordinates
        float[] transformedVerts = new float[verts.length];
        graphicsMatrix.mapPoints(transformedVerts, verts);
        
        // create a new Gouraud pattern object adn add it to the page and document
        PDFShadingGouraudFreeform shading = new PDFShadingGouraudFreeform(mode, vertexCount, transformedVerts, vertOffset, colors, colorOffset);
        
        // Add the image to the document and the page resource-reference table
        page.getPDFDocument().add(shading);
        page.addShadingResource(shading);
        
        // stroke the shading
        pw.println(shading.getName() + " sh");
        
        // for debugging shading contents
        pw.println("%" + shading.getStreamDataString());
    }
    
    
    
    /**
     * Draw the text, with origin at (x,y), using the typeface and size in the specified paint.
     * 
     * @param text
     *            The text to be drawn
     * @param x
     *            The x-coordinate of the origin of the text being drawn
     * @param y
     *            The y-coordinate of the origin of the text being drawn
     * @param paint
     *            The paint used for the text (e.g. color, size, style)
     */
    public void drawText(char[] text, int index, int count, float x, float y, Paint paint)
    {
        if ((index | count | (index + count) | (text.length - index - count)) < 0)
        {
            throw new IndexOutOfBoundsException();
        }
        
        String textString = new String(text, index, count);
        drawText(textString, x, y, paint);
    }
    
    
    
    /**
     * Draw the text, with origin at (x,y), using the typeface and size in the specified paint.
     * 
     * @param text
     *            The text to be drawn
     * @param x
     *            The x-coordinate of the origin of the text being drawn
     * @param y
     *            The y-coordinate of the origin of the text being drawn
     * @param paint
     *            The paint used for the text (e.g. color, size, style)
     */
    public void drawText(String text, float x, float y, Paint paint)
    {
        setPaint(paint);
        
        startTextBlock(x, y);
        pw.println("(" + text + ") Tj");
        endTextBlock();
    }
    
    
    
    /**
     * Draw the text, with origin at (x,y), using the typeface and size in the specified paint.
     * 
     * @param text
     *            The text to be drawn
     * @param start
     *            The index of the first character in text to draw
     * @param end
     *            (end - 1) is the index of the last character in text to draw
     * @param x
     *            The x-coordinate of the origin of the text being drawn
     * @param y
     *            The y-coordinate of the origin of the text being drawn
     * @param paint
     *            The paint used for the text (e.g. color, size, style)
     */
    public void drawText(String text, int start, int end, float x, float y, Paint paint)
    {
        if ((start | end | (end - start) | (text.length() - end)) < 0)
        {
            throw new IndexOutOfBoundsException();
        }
        
        String textString = text.substring(start, end);
        drawText(textString, x, y, paint);
    }
    
    
    
    /**
     * Draw the specified range of text, specified by start/end, with its origin at (x,y), using the typeface and size
     * in the specified paint.
     * 
     * @param text
     *            The text to be drawn
     * @param start
     *            The index of the first character in text to draw
     * @param end
     *            (end - 1) is the index of the last character in text to draw
     * @param x
     *            The x-coordinate of origin for where to draw the text
     * @param y
     *            The y-coordinate of origin for where to draw the text
     * @param paint
     *            The paint used for the text (e.g. color, size, style)
     */
    public void drawText(CharSequence text, int start, int end, float x, float y, Paint paint)
    {
        char[] buf = new char[end - start];
        TextUtils.getChars(text, start, end, buf, 0);
        drawText(buf, 0, end - start, x, y, paint);
    }
    
    
    
    /**
     * Currently unsupported; throws UnsupportedOperationException if called.
     * 
     * @param text
     *            the text to render
     * @param index
     *            the start of the text to render
     * @param count
     *            the count of chars to render
     * @param contextIndex
     *            the start of the context for shaping. Must be no greater than index.
     * @param contextCount
     *            the number of characters in the context for shaping. ContexIndex + contextCount must be no less than
     *            index + count.
     * @param x
     *            the x position at which to draw the text
     * @param y
     *            the y position at which to draw the text
     * @param dir
     *            the run direction, either {@link #DIRECTION_LTR} or {@link #DIRECTION_RTL}.
     * @param paint
     *            the paint
     * @hide
     */
    public void drawTextRun(char[] text, int index, int count, int contextIndex, int contextCount, float x, float y, int dir, Paint paint)
    {
        
        if (text == null)
        {
            throw new NullPointerException("text is null");
        }
        if (paint == null)
        {
            throw new NullPointerException("paint is null");
        }
        if ((index | count | text.length - index - count) < 0)
        {
            throw new IndexOutOfBoundsException();
        }
        if (dir != DIRECTION_LTR && dir != DIRECTION_RTL)
        {
            throw new IllegalArgumentException("unknown dir: " + dir);
        }
        
        //
        throw new UnsupportedOperationException("Operation drawTextRun not currently supported in PDFCanvas");
        
    }
    
    
    
    /**
     * Currently unsupported; throws UnsupportedOperationException if called.
     * 
     * @param text
     *            the text to render
     * @param start
     *            the start of the text to render. Data before this position can be used for shaping context.
     * @param end
     *            the end of the text to render. Data at or after this position can be used for shaping context.
     * @param x
     *            the x position at which to draw the text
     * @param y
     *            the y position at which to draw the text
     * @param dir
     *            the run direction, either 0 for LTR or 1 for RTL.
     * @param paint
     *            the paint
     * @hide
     */
    public void drawTextRun(CharSequence text, int start, int end, int contextStart, int contextEnd, float x, float y, int dir, Paint paint)
    {
        
        if (text == null)
        {
            throw new NullPointerException("text is null");
        }
        if (paint == null)
        {
            throw new NullPointerException("paint is null");
        }
        if ((start | end | end - start | text.length() - end) < 0)
        {
            throw new IndexOutOfBoundsException();
        }
        
        // int flags = dir == 0 ? 0 : 1;
        
        // To do
        throw new UnsupportedOperationException("Operation drawTextRun not currently supported in PDFCanvas");
        
    }
    
    
    
    /**
     * Draw the text in the array, with each character's origin specified by the pos array, using the typeface and size
     * in the specified Paint.
     * 
     * @param text
     *            The text to be drawn
     * @param index
     *            The index of the first character to draw
     * @param count
     *            The number of characters to draw, starting from index.
     * @param pos
     *            Array of [x,y] positions, used to position each character
     * @param paint
     *            The paint used for the text (e.g. color, size, style)
     */
    public void drawPosText(char[] text, int index, int count, float[] pos, Paint paint)
    {
        if (index < 0 || index + count > text.length || count * 2 > pos.length)
        {
            throw new IndexOutOfBoundsException();
        }
        
        setPaint(paint);
        
        for (int i = index; i < index + count; i++)
        {
            startTextBlock(pos[2 * i], pos[2 * i + 1]);
            pw.println("(" + text[i] + ") Tj");
            endTextBlock();
        }
        
    }
    
    
    
    /**
     * Draw the text in the array, with each character's origin specified by the pos array, using the typeface and size
     * in the specified Paint.
     * 
     * @param text
     *            The text to be drawn
     * @param pos
     *            Array of [x,y] positions, used to position each character
     * @param paint
     *            The paint used for the text (e.g. color, size, style)
     */
    public void drawPosText(String text, float[] pos, Paint paint)
    {
        if (text.length() * 2 > pos.length)
        {
            throw new ArrayIndexOutOfBoundsException();
        }
        
        char[] textChars = text.toCharArray();
        drawPosText(textChars, 0, textChars.length, pos, paint);
    }
    
    
    
    /**
     * Utility method to draw single character at specified position with the specified rotation, for drawing text along
     * a path
     * 
     * @param theChar
     * @param x
     * @param y
     * @param paint
     */
    private void drawChar(char theChar, float x, float y, Matrix matrix, Paint paint)
    {
        // inherit the current text font and size from setting the paint object
        pw.println("BT");
        
        // apply current text transformation matrix, but with converted position
        // (x,y) as the translation
        float[] coef = new float[9];
        matrix.getValues(coef);
        pw.println(df.format(coef[Matrix.MSCALE_X]) + " " + df.format(coef[Matrix.MSKEW_X]) + " " + df.format(coef[Matrix.MSKEW_Y]) + " " + df.format(coef[Matrix.MSCALE_Y]) + " " + cxy(x, y) + "Tm");
        pw.println("(" + theChar + ") Tj");
        
        pw.println("ET");
    }
    
    
    
    /**
     * Draw the text, with origin at (x,y), along the specified path, using the typeface and size in the specified
     * Paint.
     * 
     * @param text
     *            The text to be drawn
     * @param path
     *            The path the text should follow for its baseline
     * @param hOffset
     *            The distance along the path to add to the text's starting position
     * @param vOffset
     *            The distance above(-) or below(+) the path to position the text
     * @param paint
     *            The paint used for the text (e.g. color, size, style)
     */
    public void drawTextOnPath(char[] text, int index, int count, Path path, float hOffset, float vOffset, Paint paint)
    {
        if (index < 0 || index + count > text.length)
        {
            throw new ArrayIndexOutOfBoundsException();
        }
        
        drawTextOnPath(new String(text, index, count), path, hOffset, vOffset, paint);
    }
    
    
    
    /**
     * Draw the text, with origin at (x,y), along the specified path, using the typeface and size in the specified
     * Paint.
     * 
     * @param text
     *            The text to be drawn
     * @param path
     *            The path the text should follow for its baseline
     * @param hOffset
     *            The distance along the path to add to the text's starting position
     * @param vOffset
     *            The distance above(-) or below(+) the path to position the text
     * @param paint
     *            The paint used for the text (e.g. color, size, style)
     */
    public void drawTextOnPath(String text, Path path, float hOffset, float vOffset, Paint paint)
    {
        if (text.length() > 0)
        {
            setPaint(paint);
            
            // Use PathMeasure to get positions along the path
            PathMeasure pathMeasure = new PathMeasure(path, false);
            float pathLength = pathMeasure.getLength();
            if ((hOffset > pathLength) || (pathLength == 0))
                return;
            
            float[] textWidths = new float[text.length()];
            paint.getTextWidths(text, textWidths);
            
            float[] pos = new float[2];
            float[] tan = new float[2];
            float length = hOffset;
            
            // get starting position of text along path, and offset by
            // vOffset perpendicular to path
            pathMeasure.getPosTan(length, pos, tan);
            pos[0] += vOffset * -tan[1];
            pos[1] += vOffset * tan[0];
            
            // matrix for doing text rotation
            Matrix matrix = new Matrix();
            
            // get text rotation
            float rot = (float) (Math.acos(tan[0]) * (180 / Math.PI)) * Math.signum(tan[1]);
            matrix.setRotate(rot);
            
            for (int i = 0; i < text.length(); i++)
            {
                // draw text at pos along curve
                drawChar(text.charAt(i), pos[0], pos[1], matrix, paint);
                
                // get next position info
                length += textWidths[i];
                pathMeasure.getPosTan(length, pos, tan);
                pos[0] += vOffset * -tan[1];
                pos[1] += vOffset * tan[0];
                rot = (float) (Math.acos(tan[0]) * (180 / Math.PI)) * Math.signum(tan[1]);
                matrix.setRotate(rot);
            }
        }
    }
    
    
    
    /**
     * Currently unsupported; throws UnsupportedOperationException if called.
     * 
     * @param picture
     *            The picture to be drawn
     */
    public void drawPicture(Picture picture)
    {
        /*
         * picture.endRecording();
         * 
         * // save state save(); picture.draw(this); restore();
         */
        
        throw new UnsupportedOperationException("Operation drawPicture not supported in PDFCanvas");
    }
    
    
    
    /**
     * Currently unsupported; throws UnsupportedOperationException if called.
     */
    public void drawPicture(Picture picture, RectF dst)
    {
        /*
         * save(); translate(dst.left, dst.top); if (picture.getWidth() > 0 && picture.getHeight() > 0) {
         * scale(dst.width() / picture.getWidth(), dst.height() / picture.getHeight()); } drawPicture(picture);
         * restore();
         */
        
        throw new UnsupportedOperationException("Operation drawPicture not supported in PDFCanvas");
    }
    
    
    
    /**
     * Currently unsupported; throws UnsupportedOperationException if called.
     */
    public void drawPicture(Picture picture, Rect dst)
    {
        /*
         * save(); translate(dst.left, dst.top); if (picture.getWidth() > 0 && picture.getHeight() > 0) { scale((float)
         * dst.width() / picture.getWidth(), (float) dst.height() / picture.getHeight()); } drawPicture(picture);
         * restore();
         */
        
        throw new UnsupportedOperationException("Operation drawPicture not supported in PDFCanvas");
    }
    
    
    
    /**
     * This adds a line segment to the current path
     * 
     * @param x
     *            coordinate
     * @param y
     *            coordinate
     */
    private void lineto(float x, float y)
    {
        pw.println(cxy(x, y) + "l ");
    }
    
    
    
    /**
     * This adds a line segment to the current path
     * 
     * @param x
     *            coordinate
     * @param y
     *            coordinate
     */
    private void lineto(int x, int y)
    {
        pw.println(cxy(x, y) + "l ");
    }
    
    
    
    /**
     * This moves the current drawing point.
     * 
     * @param x
     *            coordinate
     * @param y
     *            coordinate
     */
    private void moveto(float x, float y)
    {
        pw.println(cxy(x, y) + "m ");
    }
    
    
    
    /**
     * This moves the current drawing point.
     * 
     * @param x
     *            coordinate
     * @param y
     *            coordinate
     */
    private void moveto(int x, int y)
    {
        pw.println(cxy(x, y) + "m ");
    }
    
    
    
    /**
     * Converts the Java space coordinates into pdf.
     * 
     * @param x
     *            coordinate
     * @param y
     *            coordinate
     * @return String containing the coordinates in PDF space
     */
    private String cxy(float x, float y)
    {
        float[] pts = { x, y };
        graphicsMatrix.mapPoints(pts);
        return df.format(pts[0]) + " " + df.format(pts[1]) + " ";
        
    }
    
    
    
    /**
     * Converts the Java space coordinates into pdf.
     * 
     * @param x
     *            coordinate
     * @param y
     *            coordinate
     * @return String containing the coordinates in PDF space
     */
    private String cxy(int x, int y)
    {
        return cxy((float) x, (float) y);
    }
    
    
    
    /**
     * This extension appends a Bezier curve to the path. The curve extends from the current point to (x2,y2) using the
     * current point and (x1,y1) as the Bezier control points.
     * <p>
     * The new current point is (x2,y2)
     * 
     * @param x1
     *            Second control point
     * @param y1
     *            Second control point
     * @param x2
     *            Destination point
     * @param y2
     *            Destination point
     */
    @SuppressWarnings("unused")
    private void curveto(float x1, float y1, float x2, float y2)
    {
        pw.println(cxy(x1, y1) + cxy(x2, y2) + "v");
    }
    
    
    
    /**
     * This extension appends a Bezier curve to the path. The curve extends from the current point to (x3,y3) using
     * (x1,y1) and (x2,y2) as the Bezier control points.
     * <p>
     * The new current point is (x3,y3)
     * 
     * @param x1
     *            First control point
     * @param y1
     *            First control point
     * @param x2
     *            Second control point
     * @param y2
     *            Second control point
     * @param x3
     *            Destination point
     * @param y3
     *            Destination point
     */
    private void curveto(float x1, float y1, float x2, float y2, float x3, float y3)
    {
        pw.println(cxy(x1, y1) + cxy(x2, y2) + cxy(x3, y3) + "c");
    }
    
    
    
    /**
     * This extension appends a Bezier curve to the path. The curve extends from the current point to (x2,y2) using the
     * current point and (x1,y1) as the Bezier control points.
     * <p>
     * The new current point is (x2,y2)
     * 
     * @param x1
     *            Second control point
     * @param y1
     *            Second control point
     * @param x2
     *            Destination point
     * @param y2
     *            Destination point
     */
    @SuppressWarnings("unused")
    private void curveto(int x1, int y1, int x2, int y2)
    {
        pw.println(cxy(x1, y1) + cxy(x2, y2) + "v");
    }
    
    
    
    /**
     * This extension appends a Bezier curve to the path. The curve extends from the current point to (x3,y3) using
     * (x1,y1) and (x2,y2) as the Bezier control points.
     * <p>
     * The new current point is (x3,y3)
     * 
     * @param x1
     *            First control point
     * @param y1
     *            First control point
     * @param x2
     *            Second control point
     * @param y2
     *            Second control point
     * @param x3
     *            Destination point
     * @param y3
     *            Destination point
     */
    @SuppressWarnings("unused")
    private void curveto(int x1, int y1, int x2, int y2, int x3, int y3)
    {
        pw.println(cxy(x1, y1) + cxy(x2, y2) + cxy(x3, y3) + "c");
    }
    
    
    
    /**
     * This extension appends a Bezier curve to the path. The curve extends from the current point to (x2,y2) using
     * (x1,y1) and the end point as the Bezier control points.
     * <p>
     * The new current point is (x2,y2)
     * 
     * @param x1
     *            Second control point
     * @param y1
     *            Second control point
     * @param x2
     *            Destination point
     * @param y2
     *            Destination point
     */
    @SuppressWarnings("unused")
    private void curveto2(float x1, float y1, float x2, float y2)
    {
        pw.println(cxy(x1, y1) + cxy(x2, y2) + "y");
    }
    
    
    
    /**
     * This extension appends a Bezier curve to the path. The curve extends from the current point to (x2,y2) using
     * (x1,y1) and the end point as the Bezier control points.
     * <p>
     * The new current point is (x2,y2)
     * 
     * @param x1
     *            Second control point
     * @param y1
     *            Second control point
     * @param x2
     *            Destination point
     * @param y2
     *            Destination point
     */
    @SuppressWarnings("unused")
    private void curveto2(int x1, int y1, int x2, int y2)
    {
        pw.println(cxy(x1, y1) + cxy(x2, y2) + "y");
    }
    
}
